let bsearch cmp a =
	let rec sub kara made =
		let i = ((made - kara) lsr 1) + kara in
		let x = a.(i) in
		match cmp x with
			0 -> x
			| n when n < 0 && kara <> i ->
				sub kara i
			| n when n > 0 && (i + 1) <> i->
				sub (i + 1) made
			| _ -> raise Not_found in
	sub 0 (Array.length a)

let read_index ch =
	let buf = String.create 3 in
	really_input ch buf 0 3;
	(int_of_char buf.[0])
		lor ((int_of_char buf.[1]) lsl 8)
		lor ((int_of_char buf.[2]) lsl 16)

(* 64 bit
let read_index ch =
	let buf = String.create 7 in
	really_input ch buf 0 7;
	(int_of_char buf.[0])
		lor ((int_of_char buf.[1]) lsl 8)
		lor ((int_of_char buf.[2]) lsl 16)
		lor ((int_of_char buf.[3]) lsl 24)
		lor ((int_of_char buf.[4]) lsl 32)
		lor ((int_of_char buf.[5]) lsl 40)
		lor ((int_of_char buf.[6]) lsl 48)
*)

let write_index fd i =
	let buf = String.create 3 in
	buf.[0] <- char_of_int (i land 0xff);
	buf.[1] <- char_of_int ((i lsr 8) land 0xff);
	buf.[2] <- char_of_int ((i lsr 16) land 0xff);
	if (Unix.single_write fd buf 0 3) != 3 then failwith "Fork.map"

(* 64 bit
let write_index fd i =
	let buf = String.create 7 in
	buf.[0] <- char_of_int (i land 0xff);
	buf.[1] <- char_of_int ((i lsr 8) land 0xff);
	buf.[2] <- char_of_int ((i lsr 16) land 0xff);
	buf.[3] <- char_of_int ((i lsr 24) land 0xff);
	buf.[4] <- char_of_int ((i lsr 32) land 0xff);
	buf.[5] <- char_of_int ((i lsr 40) land 0xff);
	buf.[6] <- char_of_int ((i lsr 48) land 0xff);
	if (Unix.single_write fd buf 0 7) != 7 then failwith "Fork.map"
*)

let child input output f a =
	let output = (Unix.out_channel_of_descr output) in
	let input = (Unix.in_channel_of_descr input) in
	try while (true) do let i = read_index input in
		Marshal.to_channel output (f a.(i)) [Marshal.Closures];
		flush output done
	with End_of_file -> exit 0

type child = {
	infd: Unix.file_descr;
	outfd: Unix.file_descr;
	pid: int;
	mutable i: int;
	mutable buffer: string;
	mutable have: int;
	mutable want: int }

let map nprocs size (f: ('a -> 'b)) (a: 'a array) = 
	let nprocs = min nprocs (Array.length a) in
	let size = max size (Marshal.header_size + 1) in
	let fds = ref [] in
	let children = Array.init nprocs (fun i ->
		let from_child, to_parent = Unix.pipe () in
		let from_parent, to_child = Unix.pipe () in
		match Unix.fork () with
			0 ->
				child from_parent to_parent f a;
				assert false
			| pid ->
				write_index to_child i;
				fds := from_child :: !fds;
				{
					infd = from_child;
					outfd = to_child;
					i = i;
					buffer = String.create size;
					have = 0;
					want = 0;
					pid = pid }) in
	Array.fast_sort (fun a b -> compare a.infd b.infd) children;
	let child_of_infd fd =
		bsearch (fun x -> compare fd x.infd) children in
	let results = Array.make (Array.length a) (Obj.magic ()) in
	let i = ref nprocs in
	let result_process child =
		results.(child.i) <- (Marshal.from_string child.buffer 0: 'b);
		if !i < Array.length a then (
			write_index child.outfd !i;
			child.i <- !i;
			i := !i + 1;
			child.have <- 0)
		else (
			Unix.close child.outfd;
			Unix.close child.infd;
			fds := List.filter ((<>) child.infd) !fds;
			Unix.kill child.pid Sys.sigkill) in
	let header_process child =
		let n = Unix.read child.infd child.buffer child.have
			((String.length child.buffer) - child.have) in
		child.have <- n + child.have;
		if child.have >= Marshal.header_size then (
			child.want <- Marshal.total_size child.buffer 0;
			if child.want = child.have then result_process child
			else if (String.length child.buffer) < child.want
					then (
				let tmp = String.create child.want in
				String.blit child.buffer 0 tmp 0 child.have;
				child.buffer <- tmp)) in
	let data_process child =
		let n = Unix.read child.infd child.buffer child.have
			(child.want - child.have) in
		child.have <- n + child.have;
		if child.have = child.want then result_process child in
	while (!fds != []) do
		let ready, _, _ = Unix.select !fds [] [] 0.0 in
		List.iter
			(fun fd ->
				let child = child_of_infd fd in
				if child.have < Marshal.header_size then
					header_process child
				else data_process child)
			ready done;
	for i = 0 to (nprocs - 1) do ignore (Unix.waitpid [] children.(i).pid) done;
	results
